{% extends "base.html" %}

{% block content %}
  <h1>Please complete the fields below</h1>
  <form action="" method="post" name="login">
      {{ form.hidden_tag() }}

      <p>Enter Organisation Name:{{ form.orgName(size=20) }}</p><br>
      <p>Enter Organisation Email:{{ form.orgEmailid(size=20) }}</p><br>

      <p><input type="submit" value="Confirm"></p>
  </form>
{% endblock %}
