{% extends "base.html" %}
{% from "macros.html" import render_field %}

{% block content %}
  <form method="POST">
    <div class="row">
      <div class="col-md-9 col-sm-8 col-xs-12">
        <h1>{{ form.name.data }}</h1>
      </div>
      <div class="col-md-9 col-sm-8 col-xs-12">
        {{ form.hidden_tag() }}
        {{ render_field(form.client_id) }}
        {{ render_field(form.client_secret) }}
        <div class="form-group">
          {{ form.revoke }}{{ form.reset }}
        </div>
      </div>
    </div>
    <div class="row">
      <div class="col-md-4 col-sm-4 col-xs-4">
        {{ render_field(form.name) }}
        {{ render_field(form.homepage_url) }}
      </div>
      <div class="col-md-6 col-sm-6 col-xs-12">
        {{ render_field(form.description) }}
        {{ render_field(form.callback_urls) }}
        <div class="form-group">
          {{ form.update }}{{ form.delete }}
        </div>
      </div>
    </div>
  </form>
{% endblock %}
